#include "max-in-sub-array.h"

#include<iostream>
#include<cassert>
#include<cstdlib>

using namespace std;

template<typename T>
void make_random_vector(vector<T>& v, size_t size) {
  v.resize(size);
  for (size_t i = 0; i < size; i++)
    v[i] = static_cast<T>(rand());
}

template<typename T>
T make_random_integer(T a, T b)
{
  assert(a <= b);
  T result(a + rand() % (b - a + 1));
  assert(result >= a);
  assert(result <= b);
  return result;
}

template<typename T>
void verify(buffer<T>& b, size_t first, size_t count)
{
  T max = 0;
  for (size_t i = 0; i < count; i++)
    if (b.element(first + i) > max)
      max = b.element(first + i);
  assert(b.validate(max, first, count));
  assert(b.validate(T(-1), first, count));
  if (T(max+1)) assert(b.validate(T(max+1), first, count));
  if (max>T(1)) assert(!b.validate(max-1, first, count));
}

int main()
{
  typedef unsigned short T;
  vector<T> v, vsub;
  buffer<T> b, bsub;

  for (int i = 0; i < 10000; i++) {
    size_t size = make_random_integer<size_t>(1, 3000);
    make_random_vector(v, size);
    b.bufferData(v);
    verify(b, 0, size);
    for (size_t j = 0; j < size; j++)
      assert(v[j] == b.element(j));

    for (size_t k = 0; k < 10; k++) {
      size_t offset = make_random_integer<size_t>(0, size);
      verify(b, offset, size - offset);
      size_t subsize = make_random_integer<size_t>(0, size - offset);
      verify(b, offset, subsize);

      make_random_vector(vsub, subsize);
      bsub.bufferData(vsub);

      b.bufferSubData(offset, vsub);
      verify(b, 0, size);
      verify(b, offset, subsize);
      for (size_t j = 0; j < 100; j++) {
        size_t offset2 = make_random_integer<size_t>(0, size);
        size_t subsize2 = make_random_integer<size_t>(0, size - offset2);
        verify(b, offset2, subsize2);
      }
    }
  }
  
  cout << "success" << endl;
}
