#pragma once

#include<vector>
#include<stdint.h>
#include <cstdlib>
#include<iostream>
using namespace std;
using std::size_t;

uint32_t log2int(uint32_t v) {

  const int MultiplyDeBruijnBitPosition[32] =
  {
    0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
    8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31
  };

  v |= v >> 1; // first round down to one less than a power of 2
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;

  return MultiplyDeBruijnBitPosition[(uint32_t)(v * 0x07C4ACDDU) >> 27];
}

size_t upper_power_of_two(size_t v)
{
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    return v;
}

enum { SMART = 0, NAIVE = 1 };

template<typename T, int options = SMART>
struct buffer {

  std::vector<T> data;
  size_t potsize;

  buffer() : potsize(0) {}

  const T& element(size_t i) const { return data[storage(i)]; }
  T& element(size_t i) { return data[storage(i)]; }

  bool validate(T max_allowed, size_t first, size_t count) {
    if (options == NAIVE) {
      size_t start = storage(first);
      size_t last = storage(first + count - 1);
      for (size_t i = start; i <= last; i++)
        if (data[i] > max_allowed)
          return false;
    } else {
      if (data[1] <= max_allowed)
        return true;
      size_t start = storage(first);
      size_t last = storage(first + count - 1);
      if (data[start >> (log2int(start ^ last) + 1)] <= max_allowed)
        return true;
      while (true) {
        if (last <= start) {
          if (last < start)
            return true;
          return data[start] <= max_allowed;
        }
        if (is_right(start)) {
          if (data[start] > max_allowed)
            return false;
          start = right_neighbour(start);
        }
        if (!is_right(last)) {
          if (data[last] > max_allowed)
            return false;
          last = left_neighbour(last);
        }
        start = parent(start);
        last = parent(last);
      }
    }
    return true;
  }

  void bufferData(const std::vector<T>& d) {
    bufferData(d.data(), d.size());
  }

  void bufferData(const T* ptr, size_t size) {
    potsize = upper_power_of_two(size);
    data.resize(2 * potsize);
    bufferSubData(0, ptr, size);
  }

  void bufferSubData(size_t pos, const std::vector<T>& d) {
    bufferSubData(pos, d.data(), d.size());
  }

  void bufferSubData(size_t pos, const T* ptr, size_t size) {
    for (size_t i = 0; i < size; i++)
      element(pos + i) = ptr[i];

    update_tree(storage(pos), storage(pos + size - 1));
  }

private:

  size_t storage(size_t user_index) const {
    return potsize + user_index;
  }

  static size_t parent(size_t i) {
    return i >> 1;
  }

  static bool is_right(size_t i) {
    return i & 1;
  }

  static bool is_left(size_t i) {
    return !is_right(i);
  }

  static size_t left_child(size_t i) {
    return i << 1;
  }

  static size_t right_child(size_t i) {
    return sibling(left_child(i));
  }

  static size_t left_neighbour(size_t i) {
    return i - 1;
  }

  static size_t right_neighbour(size_t i) {
    return i + 1;
  }

  static size_t sibling(size_t i) {
    return i ^ 1;
  }

  void update_tree(size_t start, size_t last) {
    if (options == NAIVE)
      return;
    if (!potsize)
      return;

    while (true) {
      if (start == last) {
        size_t startparent = parent(start);
        while (startparent) {
          data[startparent] = std::max(data[start], data[sibling(start)]);
          start = startparent;
          startparent = parent(start);
        }
        break;
      }
      start = parent(start);
      last = parent(last);
      size_t child = left_child(start);
      size_t parent = start;

#define COMPUTE_MAX_OF_PAIR \
          { \
            T a = data[child]; \
            child = right_neighbour(child); \
            T b = data[child]; \
            child = right_neighbour(child); \
            data[parent] = std::max(a, b); \
            parent = right_neighbour(parent); \
          }

      while (parent + 7 <= last)
      {
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
        COMPUTE_MAX_OF_PAIR
      }
      while (parent <= last)
      {
        COMPUTE_MAX_OF_PAIR
      }
    }
  }

};
